<?php

namespace Drupal\support_chat\Controller;

use Drupal\Core\Queue\QueueDatabaseFactory;
use Drupal\node\Entity\Node;
use Drupal\user\Entity\User;
use Drupal\Core\Controller\ControllerBase;
use Drupal\encrypt\Entity\EncryptionProfile;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\JsonResponse;
use Symfony\Component\Validator\Constraints as Assert;
use Symfony\Component\Validator\Validation;

/**
 * Constants for names of database tables.
 */
define("CHAT_TABLE", "support_chat_chats");
define("MSG_TABLE", "support_chat_messages");
define("QUEUE_TABLE", "support_chat_queue");

class SupportChatController extends ControllerBase {
    private $config;

    public function __construct() {
        $this->config = \Drupal::config('support_chat.settings');
    }
    // CHAT API FUNCTIONS
    public function session_api() {
        return new JsonResponse(['sid' => $this->get_session()]);
    }

    public function poll(String $messageID) {
        $currentUser = (int)\Drupal::currentUser()->id();
        if ($currentUser === 0) {
            $currentUser = $this->get_session();
        }

        $chats = \Drupal::database()->query('SELECT customer, support_rep, id FROM '. CHAT_TABLE .' WHERE active = 1 AND (customer = :currentUser OR support_rep = :currentUser)', [
            ':currentUser' => $currentUser
        ])->fetchAll();

        if (count($chats) < 1) {
            return new JsonResponse(['chats' => [], 'users' => (object)[]]);
        }

        // Getting a unique list of users listed in the chat table.
        $chat_users = array_unique(array_map(function($chat) {
            return $chat->customer == \Drupal::currentUser()->id() || \Drupal::currentUser()->id() == 0 ? $chat->support_rep : $chat->customer;
        }, $chats));
        $chat_ids = array_map(function($chat) {
            return $chat->id;
        }, $chats);

        $messages = \Drupal::database()->query('SELECT * FROM '. MSG_TABLE .' WHERE chat_id IN (:chats[]) AND id > :messageID', [
            ':chats[]' => $chat_ids,
            ':messageID' => $messageID
        ])->fetchAll();

        $senders = \Drupal::database()->query('SELECT DISTINCT sender FROM '. MSG_TABLE .' WHERE chat_id IN (:chats[]) AND sender != :currentUser AND sender NOT IN (:chat_users[])', [
            ':chats[]' => $chat_ids,
            ':currentUser' => $currentUser,
            ':chat_users[]' => $chat_users
        ])->fetchAll();

        $senders = array_map(function($sender) {
            return $sender->sender;
        }, $senders);
        $users = array_merge($chat_users, $senders);

        // Build assoc array to be output as JSON object with user IDs as keys.
        $users_with_names = [];
        foreach ($users as $user) {
            $users_with_names[$user] = preg_match("/[a-z]+/i", $user) > 0 ? 'Anonymous ' . substr($user, 0, 5) : User::load(intval($user))->getUsername();
        }

        $chat_output = [];
        // Supply an object to map chat IDs to an index for easy insertion of messages.
        $chat_id_map = [];
        for ($i = 0; $i < count($chats); $i++) {
            $chat = $chats[$i];
            $chat_id_map[$chat->id] = $i;
            array_push($chat_output, (object) array("id" => $chat->id, "other_user_id" => $chat->customer == $currentUser ? $chat->support_rep : $chat->customer, "messages" => []));
        }

        foreach ($messages as $message) {
            $idx = $chat_id_map[$message->chat_id];
            $content = $message->content;

            if ($this->config->get('support_chat_use_encryption') == 1) {
                $encryption_profile = $this->config->get('support_chat_encryption_profile');
                $encryption_profile = $encryption_profile != NULL ? EncryptionProfile::load($encryption_profile) : NULL;

                if ($encryption_profile !== NULL) {
                    $content = \Drupal::service('encryption')->decrypt($content, $encryption_profile);
                }
            }

            $message->content = $this->filterHTML($content);
            array_push($chat_output[$idx]->messages, $message);
        }

        return new JsonResponse(['chats' => $chat_output, 'users' => $users_with_names]);
    }

    public function send_message(Request $request) {
        $schema_constraint = new Assert\Collection([
            'fields' => [
                'chat_id' => new Assert\Type(array('type'=>'integer')),
                'content' => new Assert\Type(array('type'=>'string')),
                'sender' => new Assert\Type(array('type'=>'string'))
            ]
        ]);
        $post = (array)json_decode($request->getContent());
        $isValid = $this->isValid($post, $schema_constraint);

        if (!$isValid) {
            return new JsonResponse(['message' => "Sending a chat message requires these fields: chat_id, content, sender."], 422);
        }

        $chat_id = $post['chat_id'];
        $content = $post['content'];
        $sender = $post['sender'];

        // Verify that the sender is the current Drupal user.
        if (!$this->verifyLoggedInUser($sender)) {
            return new JsonResponse(['message' => "Sender ID does not correspond to logged in user."], 403);
        }

        // Verify that a chat is active and that sender is a member of the chat.
        $chat = \Drupal::database()->query('SELECT active FROM '. CHAT_TABLE .' WHERE id = :chat_id AND (customer = :sender OR support_rep = :sender)', [
            ':chat_id' => $chat_id,
            ':sender' => $sender
        ])->fetchField();

        if ($chat === false) {
            return new JsonResponse(['message' => "Sender is not a member of this chat."], 400);
        } else if ($chat != '1') {
            return new JsonResponse(['message' => "Chat is inactive."], 400);
        }

        if ($this->config->get('support_chat_use_encryption') == 1) {
            $encryption_profile = $this->config->get('support_chat_encryption_profile');
            $encryption_profile = $encryption_profile != NULL ? EncryptionProfile::load($encryption_profile) : NULL;

            if ($encryption_profile !== NULL) {
                $content = \Drupal::service('encryption')->encrypt($content, $encryption_profile);
            }
        }

        try {
            // Should return the message ID.
            $results = \Drupal::database()->insert(MSG_TABLE)
                ->fields([
                    'content' => $content,
                    'sender' => $sender,
                    'chat_id' => $chat_id,
                    'timestamp' => time()
                ])
                ->execute();

            return (int)$results > 0
                ? new JsonResponse(['message' => "Success"])
                : new JsonResponse(['message' => "Error inserting message into database."], 500);
        } catch (Exception $e) {
            \Drupal::logger('chat')->error($e->getMessage());
            return new JsonResponse(['message' => "Something went wrong.", 'error' => $e->getMessage()], 503);
        }
    }

    public function enqueue(Request $request) {
        if ($this->isQueueOpen() === false) {
            return new JsonResponse(['message' => 'Queue closed', 'isOpen' => false]);
        }

        $answer_required = $this->config->get('support_chat_screening_required') === 1;
        $schema_constraint = $answer_required === true
        ? new Assert\Collection([
            'fields' => [
                'customer' => new Assert\Type(array('type'=>'string')),
                'screening_response' => new Assert\Type(array('type'=>'string'))
            ]
          ])
        : new Assert\Collection([
            'fields' => [
                'customer' => new Assert\Type(array('type'=>'string'))
            ]
          ]);

        $user = \Drupal::currentUser();
        $post = (array)json_decode($request->getContent());
        $isValid = $this->isValid($post, $schema_constraint);

        if (!$isValid) {
            return new JsonResponse(['message' =>
                $answer_required === true
                ? "Enqueue requires a valid customer id (customer) and response to the screening question (screening_response)."
                : "Enqueue requires a valid customer id (customer)."],
            422);
        }

        $customer = $post['customer'] ?? $user->id();
        $screening_response = $post['screening_response'] ?? NULL;

        if (!$this->verifyLoggedInUser($customer)) {
            return new JsonResponse(['message' => "Customer does not correspond to logged in user."], 403);
        }

        $fields = [
            'customer' => $customer,
            'screening_response' => $screening_response,
            'timestamp' => time()
        ];

        try {
            $results = \Drupal::database()->insert(QUEUE_TABLE)->fields($fields)->execute();
            return (int)$results > 0 ? new JsonResponse(['message' => "Success"])
                : new JsonResponse(['message' => "Error inserting into queue table."], 500);
        } catch (\Exception $e) {
            return new JsonResponse(['message' => "Something went wrong.", 'error' => $e->getMessage(), 500]);
        }
    }

    public function dequeue(Request $request) {
        $schema_constraint = new Assert\Collection([
            'fields' => [
                'customer' => new Assert\Type(array('type'=>'string'))
            ]
        ]);

        $post = (array)json_decode($request->getContent());
        $isValid = $this->isValid($post, $schema_constraint);

        if (!$isValid) {
            return new JsonResponse(['message' => "Dequeue requires a valid customer id (customer)."], 422);
        }

        try {
            $results = \Drupal::database()->delete(QUEUE_TABLE)
                ->condition('customer', $post['customer'])->execute();

            return (int)$results > 0 ? new JsonResponse(['message' => "Success"])
                : new JsonResponse(['message' => "Error removing from queue table."], 500);

        } catch (\Exception $e) {
            return new JsonResponse(['message' => "Something went wrong.", 'error' => $e->getMessage(), 500]);
        }
    }

    public function queue_status() {
        $open = $this->isQueueOpen();
        return new JsonResponse(['open' => $open]); 
    }

    public function enumerate_queue() {
        try {
            $records = \Drupal::database()->select(QUEUE_TABLE, 'q')
                ->fields('q', ['customer', 'screening_response', 'timestamp'])->execute()->fetchAll();

            usort($records, function($a, $b) { return strcmp($b->timestamp, $a->timestamp);});

            for ($i = 0; $i < count($records); $i++) {
                $customer = $records[$i]->customer;
                $records[$i]->customer_username = preg_match("/[a-z]+/i", $customer) > 0 ? 'Anonymous ' . substr($customer, 0, 5) : User::load(intval($customer))->getUsername();
                $records[$i]->screening_response = $this->filterHTML($records[$i]->screening_response);
            }

            return new JsonResponse(['records' => $records]);
        } catch (\Exception $e) {
            return new JsonResponse(['message' => "Error", 500]);
        }
    }

    public function queue_position() {
        $currentUser = (int)\Drupal::currentUser()->id();
        if ($currentUser === 0) {
            $currentUser = $this->get_session();
        }

        try {
            $queueID = \Drupal::database()->query('SELECT id FROM '. QUEUE_TABLE .' WHERE customer = :currentUser', [
                ':currentUser' => $currentUser
            ])->fetchField();

            if ($queueID === false) {
                return new JsonResponse(['position' => NULL]);
            }

            $position = \Drupal::database()->query('SELECT COUNT(*) FROM '. QUEUE_TABLE .' WHERE id < :queueID', [
                ':queueID' => $queueID
            ])->fetchField() + 1;

            return new JsonResponse(['position' => $position]);
        } catch (\Exception $e) {
            return new JsonResponse(['message' => "Error", 500]);
        }
    }

    public function start_chat(Request $request) {
        $schema_constraint = new Assert\Collection([
            'fields' => [
                'customer' => new Assert\Type(array('type'=>'string')),
                'support_rep' => new Assert\Type(array('type'=>'integer'))
            ]
        ]);
        $post = (array)json_decode($request->getContent());
        $isValid = $this->isValid($post, $schema_constraint);

        if (!$isValid) {
            return new JsonResponse(['message' => "Starting a chat requires a customer user ID (or session) and a support rep user ID."], 422);
        }

        if (!$this->verifyLoggedInUser($post['support_rep'])) {
            return new JsonResponse(['message' => "Support rep ID does not correspond to logged in user."], 403);
        }

        try {
            // Should return the message ID.
            $results = \Drupal::database()->insert(CHAT_TABLE)
                ->fields([
                    'customer' => $post['customer'],
                    'support_rep' => $post['support_rep'],
                    'active' => 1
                ])
                ->execute();

            return (int)$results > 0
                ? new JsonResponse(['chat_id' => (int)$results])
                : new JsonResponse(['message' => "Error creating chat in database."], 500);
        } catch (Exception $e) {
            \Drupal::logger('chat')->error($e->getMessage());
            return new JsonResponse(['message' => "Something went wrong.", 'error' => $e->getMessage()], 503);
        }
    }

    public function close_chat(Request $request) {
        $schema_constraint = new Assert\Collection([
            'fields' => [
                'chat_id' => new Assert\Type(array('type'=>'integer')),
            ]
        ]);
        $put = (array)json_decode($request->getContent());
        $isValid = $this->isValid($put, $schema_constraint);

        if (!$isValid) {
            return new JsonResponse(['message' => "Please supply the chat ID to close the chat."], 422);
        }

        $chat_id = $put['chat_id'];

        $closer = \Drupal::currentUser();
        if (!$closer->hasPermission('support rep')) {
            $currentUser = (int)$closer->id();
            if ($currentUser === 0) {
                $currentUser = $this->get_session();
            }
            $chat = \Drupal::database()->query('SELECT active FROM '. CHAT_TABLE .' WHERE id = :chat_id AND (customer = :currentUser OR support_rep = :currentUser)', [
                ':chat_id' => $chat_id,
                ':currentUser' => $currentUser
            ])->fetchField();

            if ($chat === false) {
                return new JsonResponse(['message' => "Sender is not a member of this chat and does not have support rep permissions."], 400);
            }
        }

        try {
            if ($this->config->get('support_chat_delete_closed_chats') === 1) {
                $results = \Drupal::database()->delete(CHAT_TABLE)
                    ->condition('id', $put['chat_id'])->execute();

                if ((int)$results > 0) {
                    $chat = \Drupal::database()->query('SELECT content FROM '. MSG_TABLE .' WHERE chat_id = :chat_id', [
                        ':chat_id' => $chat_id
                    ])->fetchField();

                    if ($chat === false) {
                        return new JsonResponse(['message' => "Successfully closed chat " . $put['chat_id']]);
                    }

                    $results = \Drupal::database()->delete(MSG_TABLE)
                        ->condition('chat_id', $put['chat_id'])->execute();

                    return (int)$results > 0
                        ? new JsonResponse(['chat_id' => "Successfully closed chat " . $put['chat_id']])
                        : new JsonResponse(['message' => "Error deleting messages in database."], 500);
                } else {
                    return new JsonResponse(['message' => "Error deleting chat in database."], 500);
                }
            } else {
                // Should return the message ID.
                $results = \Drupal::database()->update(CHAT_TABLE)
                    ->fields([
                        'active' => 0
                    ])
                    ->condition('id', $put['chat_id'])
                    ->execute();

                return (int)$results > 0
                    ? new JsonResponse(['chat_id' => "Successfully closed chat " . $put['chat_id']])
                    : new JsonResponse(['message' => "Error updating chat in database."], 500);
            }
        } catch (Exception $e) {
            \Drupal::logger('chat')->error($e->getMessage());
            return new JsonResponse(['message' => "Something went wrong.", 'error' => $e->getMessage()], 503);
        }
    }

    public function transfer_chat(Request $request) {
        $schema_constraint = new Assert\Collection([
            'fields' => [
                'chat_id' => new Assert\Type(array('type'=>'integer')),
                'support_rep' => new Assert\Type(array('type'=>'integer'))
            ]
        ]);
        $put = (array)json_decode($request->getContent());
        $isValid = $this->isValid($put, $schema_constraint);

        if (!$isValid) {
            return new JsonResponse(['message' => "Transferring a chat requires these fields: chat_id, support_rep"], 422);
        }

        // Verify that the account corresponding to the provided ID has the support rep permission.
        $dest_user = User::load($put['support_rep']);
        if ($dest_user === null) {
            return new JsonResponse(['message' => "User with provided ID not found."], 400);
        }
        if (!$dest_user->hasPermission('support rep')) {
            return new JsonResponse(['message' => "User with provided ID does not have the support rep permission."], 400);
        }

        try {
            // Should return the message ID.
            $results = \Drupal::database()->update(CHAT_TABLE)
                ->fields([
                    'support_rep' => $put['support_rep'],
                    'active' => 1
                ])
                ->condition('id', $put['chat_id'])
                ->execute();

            return (int)$results > 0
                ? new JsonResponse(['chat_id' => "Successfully transferred chat " . $put['chat_id']])
                : new JsonResponse(['message' => "Error updating chat in database."], 500);
        } catch (Exception $e) {
            \Drupal::logger('chat')->error($e->getMessage());
            return new JsonResponse(['message' => "Something went wrong.", 'error' => $e->getMessage()], 503);
        }
    }

    public function toggle_availability(String $availability) {
        $currentUser = (int)\Drupal::currentUser()->id();
        if ($currentUser === 0) {
            return new JsonResponse(['message' => "Cannot change availability of anonymous users."], 400);
        }
        $user = User::load($currentUser);

        switch ($availability) {
            case "available":
                $user->set('field_support_chat_availability', true);
                break;
            case "unavailable":
                $user->set('field_support_chat_availability', false);
                break;
            default:
                return new JsonResponse(['message' => `Invalid option. Should be "available" or "unavailable".`], 400);
        }

        $user->save();
        return new JsonResponse(['message' => `Success`, 'available' => $user->get('field_support_chat_availability')->value]);
    }

    public function available_support_reps() {
        $support_roles = array_keys(user_role_names(FALSE, 'support rep'));

        $users = \Drupal::entityQuery('user')
            ->condition('roles', $support_roles, 'IN')
            ->condition('field_support_chat_availability', true)
            ->execute();

        $results = [];
        foreach($users as $userID) {
            $user = User::load($userID);
            $userinfo = [
                "id" => $userID,
                "username" => $user->getUsername()
            ];
            $results[$userID] = $userinfo;
        }

        return new JsonResponse(['users' => $results]);
    }

    // SUPPORTING FUNCTIONS
    private function isValid($post, $constraint) {
        $validator = Validation::createValidator();
        $error_list = $validator->validate($post, $constraint);

        if (count($error_list) < 1) {
            return true;
        }
        return false;
    }

    private function isQueueOpen() {
        return $this->config->get('support_chat_queue_open') ?? false;
    }

    private function get_session() {
        $session_manager = \Drupal::service('session_manager');
        $started = $session_manager->start();
        $sid = $session_manager->getId();
        $session = \Drupal::request()->getSession();
        $session->set("chat_sid", $sid);
        return $session->get("chat_sid");
    }

    private function verifyLoggedInUser($session_or_id) {
        $id = (int)\Drupal::currentUser()->id();
        if ($id === 0) {
            $sid = $this->get_session();
            if ($session_or_id !== $sid) {
                return false;
            }
        } else if (intval($session_or_id) !== $id) {
            return false;
        }
        return true;
    }

    private function filterHTML($string) {
        $string = str_replace('<', '&lt;', $string);
        $string = str_replace('>', '&gt;', $string);
        return filter_var($string, FILTER_SANITIZE_STRING);
    }
}
